#include "cppdefs.h"
#if defined NBQ && !defined NBQ_IJK

      subroutine mat_cont_init_nh 

!*******************************************************************
!*******************************************************************
!*******************************************************************
! Matrix CONT for continuity equation: initialization
!
! http://poc.obs-mip.fr/auclair/WOcean.fr/SNH/Restricted/NH-NBQ/Html_pages/Algebrique_Formats_CSR.htm
!
! Matrix construction:
! http://poc.obs-mip.fr/auclair/WOcean.fr/SNH/Pub/NH-NBQ/Html_maps/Schemas_Num_map.html
!
!*******************************************************************
!*******************************************************************
!*******************************************************************

      use module_nh
      use module_nbq
      implicit none

# include "param_F90.h"
# include "scalars_F90.h"
      
      integer :: i,j,k

!*******************************************************************
!     Various Initializations:
!*******************************************************************

!.....Initialize position indices, which will be used
!     as size variable afterwards:
      nzcont_nh     = 1
      nzcontz_nh     = 1

!.....initializations:
      conti_nh    = 1
      contzi_nh    = 1
      cont_nnz_nh = 1
      contz_nnz_nh = 1
      contj_nh    = 0
      contv_nh    = 0.
      contzj_nh    = 0
      contzv_nh    = 0.

!*******************************************************************
!     Continuity Equation:
!*******************************************************************

!-------------------------------------------------------------------
!     Inner domain, bottom layer: (i,j,1)
!-------------------------------------------------------------------

      do l_nh = neqq_nh(2)+1,neqq_nh(3)   

!......Equation characteristics:
       i = l2iq_nh (l_nh) 
       j = l2jq_nh (l_nh) 
       k = l2kq_nh (l_nh)      
       conti_nh (l_nh) = nzcont_nh  !! matrix line pointer
       contzi_nh (l_nh) = nzcontz_nh  !! matrix line pointer

!........u(i,j,k-1):	- 1	-

!........u(i+1,j,k-1):	- 2	-

!.......u(i,j,k):	- 3	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,1)

!.......u(i,j,k+1):	- 4	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k+1,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k+1,1)

!.......u(i+1,j,k):	- 5	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k,1)

!........u(i+1,j,k+1):	- 6	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k+1,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k+1,1)

!........v(i,j,k-1):	- 7	-

!.......v(i,j,k):	- 8	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,2)

!.......v(i,j,k+1):	- 9	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k+1,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k+1,2)

!........v(i,j+1,k-1):	- 10	-

!.......v(i,j+1,k):	- 11	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k,2)

!........v(i,j+1,k+1):	- 12	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k+1,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k+1,2)

!.......w(i,j,k):
         contzj_nh(nzcontz_nh) = ijk2lmom_nh(i,j,k,3)
         nzcontz_nh           = nzcontz_nh + mijk2lmom_nh(i,j,k,3)

!.......w(i,j,k-1):
!        contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k-1,3)
!        nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k-1,3)

!       Last point...
        cont_nnz_nh(l_nh+1)=nzcont_nh
        contz_nnz_nh(l_nh+1)=nzcontz_nh

      enddo

!-------------------------------------------------------------------
!     Inner domain, inner layers: (i,j,1<k<N)
!-------------------------------------------------------------------

      do l_nh = neqq_nh(3)+1 , neqq_nh(4)

!......Equation characteristics:
       i = l2iq_nh (l_nh) 
       j = l2jq_nh (l_nh) 
       k = l2kq_nh (l_nh)      
       conti_nh (l_nh) = nzcont_nh  !! matrix line pointer
       contzi_nh (l_nh) = nzcontz_nh  !! matrix line pointer

!........u(i,j,k-1):	- 1	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k-1,1)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k-1,1)

!........u(i+1,j,k-1):	- 2	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k-1,1)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k-1,1)

!.......u(i,j,k):	- 3	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,1)

!.......u(i,j,k+1):	- 4	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k+1,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k+1,1)

!.......u(i+1,j,k):	- 5	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k,1)

!........u(i+1,j,k+1):	- 6	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k+1,1)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k+1,1)

!........v(i,j,k-1):	- 7	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k-1,2)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k-1,2)

!.......v(i,j,k):	- 8	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,2)

!.......v(i,j,k+1):	- 9	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k+1,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k+1,2)

!........v(i,j+1,k-1):	- 10	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k-1,2)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k-1,2)

!.......v(i,j+1,k):	- 11	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k,2)

!........v(i,j+1,k+1):	- 12	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k+1,2)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k+1,2)

!.......w(i,j,k):
        contzj_nh(nzcontz_nh) = ijk2lmom_nh(i,j,k,3)
        nzcontz_nh           = nzcontz_nh + mijk2lmom_nh(i,j,k,3)

!.......w(i,j,k-1):
        contzj_nh(nzcontz_nh) = ijk2lmom_nh(i,j,k-1,3)
        nzcontz_nh           = nzcontz_nh + mijk2lmom_nh(i,j,k-1,3)

!       Last point...
        cont_nnz_nh(l_nh+1)=nzcont_nh
        contz_nnz_nh(l_nh+1)=nzcontz_nh

      enddo

!-------------------------------------------------------------------
!     Inner domain, surface layer: (i,j,N)
!-------------------------------------------------------------------

      do l_nh = neqq_nh(4)+1,neqq_nh(5)  

!......Equation characteristics:
       i = l2iq_nh (l_nh) 
       j = l2jq_nh (l_nh) 
       k = l2kq_nh (l_nh)      
       conti_nh (l_nh) = nzcont_nh  !! matrix line pointer
       contzi_nh (l_nh) = nzcontz_nh  !! matrix line pointer

!........u(i,j,k-1):	- 1	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k-1,1)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k-1,1)

!........u(i+1,j,k-1):	- 2	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k-1,1)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k-1,1)

!.......u(i,j,k):	- 3	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,1)

!.......u(i,j,k+1):	- 4	-

!.......u(i+1,j,k):	- 5	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i+1,j,k,1)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i+1,j,k,1)

!........u(i+1,j,k+1):	- 6	-

!........v(i,j,k-1):	- 7	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k-1,2)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k-1,2)

!.......v(i,j,k):	- 8	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j,k,2)

!.......v(i,j,k+1):	- 9	-

!........v(i,j+1,k-1):	- 10	-
          contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k-1,2)
          nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k-1,2)

!.......v(i,j+1,k):	- 11	-
         contj_nh(nzcont_nh) = ijk2lmom_nh(i,j+1,k,2)
         nzcont_nh           = nzcont_nh + mijk2lmom_nh(i,j+1,k,2)

!........v(i,j+1,k+1):	- 12	-

!.......w(i,j,k):	- x	-
         contzj_nh(nzcontz_nh) = ijk2lmom_nh(i,j,k,3)
         nzcontz_nh           = nzcontz_nh + mijk2lmom_nh(i,j,k,3)

!.......w(i,j,k-1):	- x	-
         contzj_nh(nzcontz_nh) = ijk2lmom_nh(i,j,k-1,3)
         nzcontz_nh           = nzcontz_nh + mijk2lmom_nh(i,j,k-1,3)

!       Last point...
        cont_nnz_nh(l_nh+1)=nzcont_nh
        contz_nnz_nh(l_nh+1)=nzcontz_nh

      enddo

!*******************************************************************
!     Last line treatment:
!*******************************************************************
      conti_nh    (neqq_nh(5)+1:neqq_nh(7)+1) = nzcont_nh
      cont_nnz_nh (neqq_nh(5)+1:neqq_nh(7)+1) = nzcont_nh
      contzi_nh    (neqq_nh(5)+1:neqq_nh(7)+1) = nzcontz_nh
      contz_nnz_nh (neqq_nh(5)+1:neqq_nh(7)+1) = nzcontz_nh

      if (ifl_nbq.eq.1.and.ifl_imp_nbq.eq.1) then
!.......................................
! Matrice schema implicit: 
!  points interieurs et points de frontiere
!.......................................

       nzcimp_nbq     = 1
       cimpj_nbq = 1
       cimpv_nbq = 0.

       do l_nh = 1,neqcont_nh

!......caracteristiques de l'equation:
       i = l2iq_nh (l_nh)
       j = l2jq_nh (l_nh)
       k = l2kq_nh (l_nh)
       cimpi_nbq(l_nh) = nzcimp_nbq

!.......point w(i,j,k):
       cimpj_nbq(nzcimp_nbq) = ijk2lmom_nh(i,j,k,3)-neqmom_nh(1)-neqmom_nh(2)
       nzcimp_nbq            = nzcimp_nbq + mijk2lmom_nh(i,j,k,3)

!.......point w(i,j,k-1):
       cimpj_nbq(nzcimp_nbq) = ijk2lmom_nh(i,j,k-1,3)-neqmom_nh(1)-neqmom_nh(2)
       nzcimp_nbq            = nzcimp_nbq + mijk2lmom_nh(i,j,k-1,3)

       enddo

      endif

!*******************************************************************
!     Last line treatment:
!*******************************************************************
      neqcimp_nbq = neqcont_nh 
      cimpi_nbq(neqcimp_nbq+1) = nzcimp_nbq

! 	write(1000,*) "neqq_nh(1:)=",neqq_nh(1:)
!       do l_nh = neqq_nh(2)+1,neqq_nh(3)   
! 	do j=conti_nh (l_nh),conti_nh (l_nh+1)-1
! 	    write(1000,*) "1 -",l_nh,j,contj_nh(j)
! 	enddo
! 	write(1000,*) 
!       enddo
!       write(1000,*) "--------"
!       do l_nh = neqq_nh(3)+1,neqq_nh(4)   
! 	do j=conti_nh (l_nh),conti_nh (l_nh+1)-1
! 	    write(1000,*) "2 -",l_nh,j,contj_nh(j)
! 	enddo
! 	write(1000,*) 
!       enddo
! 	write(1000,*)  "--------"
!        do l_nh = neqq_nh(4)+1,neqq_nh(5)   
! 	do j=conti_nh (l_nh),conti_nh (l_nh+1)-1
! 	    write(1000,*) "3 -",l_nh,j,contj_nh(j)
! 	enddo
! 	write(1000,*) 
!       enddo
     return
      end subroutine mat_cont_init_nh
#else
      subroutine mat_cont_init_nh_empty
      return
      end
#endif
